/*
 * Copyright (C) 2008-2010 Wayne Meissner
 *
 * This file is part of the JNR project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jnr.ffi.annotations;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Marks a method parameter as being pinnable.
 * <p>
 * This means the data for the parameter is not copied to/from native memory.
 * Instead, the JVM memory is locked and passed directly to the native code.
 * </p>
 * <p>
 * Because the data is passed to native code unconverted, care must be taken to
 * ensure element widths align for types wider than <code>byte</code>.
 * </p>
 * <p>
 * <b>IMPORTANT:</b> This should not be used for functions that may block on 
 * network or filesystem access such as read(2), write(2), stat(2).
 * </p>
 */
@Retention(RetentionPolicy.RUNTIME)
@Target({ ElementType.PARAMETER, ElementType.ANNOTATION_TYPE })
public @interface Pinned {
}
